from functools import reduce
from typing import Callable, Mapping, Optional, Sequence, Union

from pydantic import validate_arguments


class State:
    @validate_arguments
    def __init__(
        self,
        state_name: str,
        state_text: str,
        data_type: Callable = int,
        back_button: Optional[Union[bool, str]] = None,
        **kwargs,
    ):
        self.name = state_name
        self.msg = state_text
        self.data_type = data_type
        self.back_button = back_button
        self.kwargs = kwargs
        self.callback = None
        self.regex, self.regex_error_text = None, None
        self.action = None
        self.build, self.max_row = None, None
        self.custom = None
        self.routes = None
        (
            self.list,
            self.list_all,
            self.list_buttons,
            self.list_start,
            self.list_max_row,
            self.list_labels,
        ) = (None, None, None, None, None, None)
        self.handler, self.handler_error_text = None, None
        self.long_task = None
        self.refresh_auth = None

    def __str__(self):
        return f"State <{self.name}>"

    @validate_arguments
    def add_keyboard(
        self,
        keyboard: Union[Sequence[str], Mapping[int, str]],
        size: Optional[Sequence[int]] = None,
        max_row: int = 3,
    ):
        """Add inline keyboard handler"""
        if isinstance(keyboard, list):
            keyboard = dict(enumerate(keyboard))
        if size and sum(size) != len(keyboard):
            raise ValueError(
                f"Keyboard length ({len(keyboard)}) must be "
                "the same size as the sum of row size ({sum(size)})."
            )
        if size:
            size = reduce(
                lambda x, y: x + y,
                map(
                    lambda x: [x]
                    if x < 9
                    else [8] * (x // 8) + ([x % 8], [])[not x % 8],
                    size,
                ),
            )
        elem_n = len(keyboard)
        if not size:
            size = [max_row for _ in range(elem_n // max_row)] + (
                [r] if (r := elem_n % max_row) else []
            )
        self.callback = (keyboard, tuple(size))

    @validate_arguments
    def add_text(self, regex: str = r"^.*$", error_message: Optional[str] = None):
        """Add text input handler"""
        self.regex = regex
        self.regex_error_text = error_message

    @validate_arguments
    def add_action(self, function: Callable):
        """Add dynamic action"""
        self.action = function

    @validate_arguments
    def add_dynamic_keyboard(self, function: Callable, max_row: int = 3):
        """Add function to build a keyboard dynamically"""
        self.build = function
        self.max_row = max_row

    @validate_arguments
    def add_custom_keyboard(self, function: Callable):
        """Add function to build a custom keyboard
        must return a list of InlineKeyboardButton"""
        self.custom = function

    @validate_arguments
    def add_dynamic_routes(self, function: Callable):
        """Add function to create dynamic routes
        must return 3 value (routes,default,back)"""
        self.routes = function

    @validate_arguments
    def add_dynamic_list(
        self,
        function: Callable,
        start: int = 0,
        left_button: str = "<",
        right_button: str = ">",
        all_elements: bool = False,
        labels: Optional[Callable] = None,
        max_row: int = 4,
    ):
        """Add function to create a dynamic list with pages"""
        self.list = function
        self.list_buttons = [left_button, right_button]
        self.list_all = all_elements
        self.list_start = start
        self.list_max_row = max_row
        self.list_labels = labels

    @validate_arguments
    def add_custom_handler(
        self, handler: Callable, error_message: Optional[str] = None
    ):
        """Add function handler to handle a non-text message
        must return an hashable value used to get to next state (by routes)"""
        self.handler = handler
        self.handler_error_text = error_message

    @validate_arguments
    def set_long_task(self, text: str):
        """Add a middle message wainting for the long main task"""
        self.long_task = text

    @validate_arguments
    def add_refresh_auth(self, func: Callable):
        """Add function that return a new list of authorized users (Telegram ids)"""
        self.refresh_auth = func



from typing import Callable, Optional, Sequence, Union

from autoconv.state import State
from pydantic import validate_arguments


class Conversation:
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def __init__(
        self,
        start_state: State,
        end_state: Optional[State] = None,
        fallback_state: Optional[State] = None,
    ):
        self.start = start_state
        self.end = end_state
        self.fallback_state = fallback_state
        self.state_list = [st for st in [start_state, end_state, fallback_state] if st]
        self.routes = dict()
        self.users_list = None
        self.defaults = dict()
        self.default_func = str
        self.default_back = None

    def __str__(self):
        routes_print = "\n".join(
            [
                "  * {}:\n{}".format(
                    s,
                    "\n".join(["      {:>5} -> {}".format(v, d) for v, d in r.items()]),
                )
                for s, r in self.routes.items()
            ]
        )
        heading = f"CONVERSATION\n{self.start} ==> " + (
            f"{self.end}" if self.end else "..."
        )
        return (
            f"{heading}\n"
            f"# State list: {[str(s) for s in self.state_list]}\n"
            f"# Routes:\n{routes_print}"
        )

    def _add_states(self, states: Union[Sequence[State], State]):
        """Add states to the conversation"""
        if isinstance(states, State):
            states = (states,)
        for s in states:
            if s not in self.state_list:
                self.state_list.append(s)
            else:
                raise ValueError(f"Already exists a state with name <{s.name}>.")

    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def add_routes(
        self,
        state: State,
        routes: Optional[dict] = None,
        default: Optional[State] = None,
        back: Optional[Union[bool, State]] = None,
    ):
        """Add routes for a state"""
        if state not in self.state_list:
            self._add_states(state)
        if self.routes.get(state.name):
            self.routes.pop(state.name)
        if routes:
            for k, v in routes.items():
                if (state.callback and k not in state.callback[0]) and k in (
                    -1,
                    "BACK",
                ):
                    raise ValueError(
                        f"'{k}' it's not a possible value of {str(state)}."
                    )
                if v not in self.state_list:
                    self._add_states(v)
            self.routes.update({state.name: routes})
        if default:
            if default not in self.state_list:
                self._add_states(default)
            s.update({-1: default}) if (
                s := self.routes.get(state.name)
            ) else self.routes.update({state.name: {-1: default}})
        if back or state.back_button or self.default_back:
            if not back and (
                state.back_button
                or (state.back_button is not False and self.default_back)
            ):
                back = True
            if isinstance(back, State) and back not in self.state_list:
                self._add_states(back)
            s.update({"BACK": back}) if (
                s := self.routes.get(state.name)
            ) else self.routes.update({state.name: {"BACK": back}})

    @validate_arguments
    def get_state(self, state_name: Optional[str]):
        """Get state from states list by name"""
        if state_name == self.start.name:
            return self.start
        for state in self.state_list:
            if state.name == state_name:
                return state
        return None

    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def add_authorized_users(
        self, users_list: Sequence[int], no_auth_state: Optional[State] = None
    ):
        """Define a list of users able to access this conversation"
        and an optional fallback State"""
        self.no_auth_state = no_auth_state
        self.users_list = users_list
        self._add_states(no_auth_state)

    @validate_arguments
    def set_defaults(
        self,
        params: Optional[dict] = None,
        func: Optional[Callable] = None,
        back_button: Optional[str] = None,
    ):
        """Define default values, a function applied to text and a back button
        for every States in the conversation"""
        self.defaults = params
        self.default_func = func or str
        self.default_back = back_button



class TelegramData:
    def __init__(self, users):
        self.update = None
        self.context = None
        self.telegram_id = None
        self.udata = None
        self.sdata = None
        self.message = None
        self.exception = None
        self.users = users

    def __str__(self):
        return (
            f"UPDATE {self.update}\n"
            f"CONTEXT {self.context}\n"
            f"TELEGRAM ID {self.telegram_id}\n"
            f"USER DATA {self.udata}\n"
            f"STATES DATA {self.sdata}"
            f"AUTH USERS {self.users}"
        )

    def update_telegram_data(self, update, context):
        self.update = update
        self.context = context

    def prepare(self):
        self.telegram_id = self.update.effective_chat.id
        self.udata = self.context.user_data
        self.sdata = self.udata.get("data") if self.udata else None
        self.message = self.update.message
        return self



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="telegram-autoconv",
    version="0.0.56",
    author="Moris Doratiotto",
    author_email="moris.doratiotto@gmail.com",
    description="A python module to auto build conversation for a Telegram bot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mortafix/AutoConv-Telegram-Python",
    packages=setuptools.find_packages(),
    install_requires=["python-telegram-bot", "pydantic"],
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
    ],
    python_requires=">=3.9",
    keywords=["conversation", "generator", "telegram", "bot"],
)
